import 'dart:convert';
import 'dart:math' as math;

import 'package:flutter/material.dart';
import 'package:google_fonts/google_fonts.dart';
import 'package:intl/intl.dart';
import 'package:timeago/timeago.dart' as timeago;
import 'lat_lng.dart';
import 'place.dart';
import 'uploaded_file.dart';
import '/backend/schema/structs/index.dart';

DateTime stringToDateTime(String dateString) {
  DateTime date;
  try {
    date = DateTime.parse(dateString);
  } catch (e) {
    throw FormatException('Invalid date format');
  }

  return date;
}

String? cpfNoMask(String input) {
  return input.replaceAll(RegExp(r'[^0-9]'), '');
}

bool? isValidCPF(String cpf) {
  cpf = cpf.replaceAll(RegExp(r'[^0-9]'), '');

  // Verifica se o CPF tem 11 dígitos ou é uma sequência repetida
  if (cpf.length != 11 || RegExp(r'^(.)\1*\$').hasMatch(cpf)) {
    return false;
  }

  // Calcula o primeiro dígito verificador
  int sum = 0;
  for (int i = 0; i < 9; i++) {
    sum += int.parse(cpf[i]) * (10 - i);
  }
  int firstCheck = (sum * 10) % 11;
  if (firstCheck == 10) firstCheck = 0;
  if (firstCheck != int.parse(cpf[9])) {
    return false;
  }

  // Calcula o segundo dígito verificador
  sum = 0;
  for (int i = 0; i < 10; i++) {
    sum += int.parse(cpf[i]) * (11 - i);
  }
  int secondCheck = (sum * 10) % 11;
  if (secondCheck == 10) secondCheck = 0;
  if (secondCheck != int.parse(cpf[10])) {
    return false;
  }

  return true;
}

bool isValidEmail(String email) {
  final emailRegex =
      RegExp(r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$');
  return emailRegex.hasMatch(email);
}

dynamic dependenteToJsonLista(
  List<String>? nomedependente,
  List<String>? cpfdependente,
  List<String>? parentescodependente,
  List<String>? datanascimentodependente,
  List<String>? id,
) {
  // Verifica se todas as listas têm o mesmo tamanho e não são nulas
  if (nomedependente == null ||
      cpfdependente == null ||
      parentescodependente == null ||
      datanascimentodependente == null ||
      id == null ||
      nomedependente.length != cpfdependente.length ||
      cpfdependente.length != parentescodependente.length ||
      parentescodependente.length != datanascimentodependente.length ||
      datanascimentodependente.length != id.length) {
    throw ArgumentError(
        "Todas as listas devem ter o mesmo tamanho e não podem ser nulas.");
  }

  // Cria a lista de mapas (JSON) para cada dependente
  final List<Map<String, dynamic>> dependenteList = List.generate(
    nomedependente.length,
    (index) => {
      "id": id[index],
      "nome": nomedependente[index],
      "cpf": cpfdependente[index],
      "parentesco": parentescodependente[index],
      "data_nascimento": datanascimentodependente[index],
    },
  );

  // Retorna a lista de dependentes como um JSON
  return dependenteList;
}

int getStringLength(String string) {
  return string.length;
}

String? formatDateTimeToDateTime(String? inputDateTime) {
  if (inputDateTime == null || inputDateTime.isEmpty) {
    return null; // Retorna nulo se a entrada for inválida
  }

  try {
    // Converte a string de entrada para um objeto DateTime
    DateTime parsedDateTime = DateTime.parse(inputDateTime);

    // Formata a data no formato dd/MM/yyyy
    final formattedDate = DateFormat('dd/MM/yyyy HH:mm').format(parsedDateTime);

    return formattedDate;
  } catch (e) {
    // Retorna nulo em caso de erro de parsing
    return null;
  }
}

bool hasUpperCase(String string) {
  return string.contains(RegExp(r'[A-Z]'));
}

bool hasLowerCase(String string) {
  return string.contains(RegExp(r'[a-z]'));
}

bool hasNumbers(String string) {
  return string.contains(RegExp(r'[0-9]'));
}

bool hasSymbols(String string) {
  return string.contains(RegExp(r'[!@#\\$%^&*(),.?":{}|<>]'));
}

dynamic decodeJWT(String token) {
  try {
    final parts = token.split('.');
    if (parts.length != 3) {
      throw Exception('Token inválido');
    }

    final payload = parts[1];
    final normalized = base64Url.normalize(payload);
    final decoded = utf8.decode(base64Url.decode(normalized));

    return jsonDecode(decoded) as Map<String, dynamic>;
  } catch (e) {
    print('Erro ao decodificar o JWT: \$e');
    return null;
  }
}

String removeCpfMask(String cpf) {
  return cpf.replaceAll(RegExp(r'[^0-9]'), '');
}

String formatDateToDB(String date) {
  try {
    // Divida a string pelo separador "/"
    final parts = date.split('/');

    if (parts.length != 3) {
      throw FormatException('Formato de data inválido');
    }

    // Extraia dia, mês e ano
    final day = int.parse(parts[0]);
    final month = int.parse(parts[1]);
    final year = int.parse(parts[2]);

    // Validação básica do ano
    final currentYear = DateTime.now().year;
    if (year < 1900 || year > currentYear) {
      throw FormatException(
          'Ano inválido. Deve estar entre 1900 e $currentYear.');
    }

    // Tente criar uma data para verificar validade
    final parsedDate = DateTime(year, month, day);

    // Verificar se o mês e o dia são coerentes
    if (parsedDate.month != month || parsedDate.day != day) {
      throw FormatException('Dia ou mês inválido.');
    }

    // Retorne no formato yyyy-MM-dd
    return DateFormat('yyyy-MM-dd').format(parsedDate);
  } catch (e) {
    return 'Erro: $e';
  }
}

dynamic montaJsonAlteraRepresentante(
  String? tipoid,
  String? usuarioid,
  bool? ciente,
  bool? confirmacao,
  String? categoria,
) {
  final Map<String, dynamic> jsonData = {
    "tipo_id": int.tryParse(tipoid ?? '0') ?? 0, // Converte para inteiro
    "usuario_id": int.tryParse(usuarioid ?? '0') ?? 0, // Converte para inteiro
    "ciente": ciente?.toString() ?? "False", // Converte o bool para string
    "confirmacao":
        confirmacao?.toString() ?? "False", // Converte o bool para string
    "categoria":
        categoria ?? "Sem Categoria", // Valor padrão se categoria for nulo
  };
  return jsonData;
}

bool containsString(
  String procurar,
  String em,
) {
  return em.contains(procurar);
}

dynamic montaJsonRepresentantes(
  String? txtContatoPrincipal,
  String? txtOutroContato,
  String? txtCEP,
  String? txtLogradouro,
  String? txtNumero,
  String? txtQuadra,
  String? txtLote,
  String? txtBairro,
  String? txtCidadeEstado,
  String? txtPontoReferencia,
) {
  final Map<String, dynamic> enderecoData = {
    "contato1": txtContatoPrincipal,
    "contato2": txtOutroContato,
    "cep": txtCEP,
    "logradouro": txtLogradouro,
    "numero": txtNumero,
    "quadra": txtQuadra,
    "lote": txtLote,
    "bairro": txtBairro,
    "cidade_estado": txtCidadeEstado,
    "ponto_referencia": txtPontoReferencia,
  };
  return enderecoData;
}

dynamic montaJsonAlteraSolicitacao(
  String? tipoid,
  String? usuarioid,
  bool? ciente,
  bool? confirmacao,
  String? categoria,
  String? solicitacaoid,
) {
  final Map<String, dynamic> jsonData = {
    "tipo_id": int.tryParse(tipoid ?? '0') ?? 0, // Converte para inteiro
    "usuario_id": int.tryParse(usuarioid ?? '0') ?? 0, // Converte para inteiro
    "ciente": ciente?.toString() ?? "False", // Converte o bool para string
    "confirmacao":
        confirmacao?.toString() ?? "False", // Converte o bool para string
    "categoria":
        categoria ?? "Sem Categoria", // Valor padrão se categoria for nulo
    "solicitacao_id": solicitacaoid?.toString() ?? "0", // Converte para inteiro
  };
  return jsonData;
}

dynamic montaJsonIncluirDependente(
  String? tipoid,
  String? usuarioid,
  bool? ciente,
  bool? confirmacao,
  String? categoria,
) {
  final Map<String, dynamic> jsonData = {
    "tipo_id": int.tryParse(tipoid ?? '0') ?? 0, // Converte para inteiro
    "usuario_id": int.tryParse(usuarioid ?? '0') ?? 0, // Converte para inteiro
    "ciente": ciente?.toString() ?? "False", // Converte o bool para string
    "confirmacao":
        confirmacao?.toString() ?? "False", // Converte o bool para string
    "categoria":
        categoria ?? "Sem Categoria", // Valor padrão se categoria for nulo
  };
  return jsonData;
}

dynamic montaJsonAlteraDependente(
  String? tipoid,
  String? usuarioid,
  bool? ciente,
  bool? confirmacao,
  String? categoria,
) {
  final Map<String, dynamic> jsonData = {
    "tipo_id": int.tryParse(tipoid ?? '0') ?? 0, // Converte para inteiro
    "usuario_id": int.tryParse(usuarioid ?? '0') ?? 0, // Converte para inteiro
    "ciente": ciente?.toString() ?? "False", // Converte o bool para string
    "confirmacao":
        confirmacao?.toString() ?? "False", // Converte o bool para string
    "categoria":
        categoria ?? "Sem Categoria", // Valor padrão se categoria for nulo
  };
  return jsonData;
}

dynamic montaJsonDependentes(
  String? txtNomeDependente,
  String? txtCpfDependente,
  String? txtParentescoDependente,
  String? txtDataNascimentoDependente,
) {
  final Map<String, dynamic> enderecoData = {
    "nome": txtNomeDependente,
    "cpf": txtCpfDependente,
    "parentesco": txtParentescoDependente,
    "data_nascimento": txtDataNascimentoDependente,
  };
  return enderecoData;
}

dynamic montaJsonExcluirDependente(
  String? tipoid,
  String? usuarioid,
  bool? ciente,
  bool? confirmacao,
  String? categoria,
) {
  final Map<String, dynamic> jsonData = {
    "tipo_id": int.tryParse(tipoid ?? '0') ?? 0, // Converte para inteiro
    "usuario_id": int.tryParse(usuarioid ?? '0') ?? 0, // Converte para inteiro
    "ciente": ciente?.toString() ?? "False", // Converte o bool para string
    "confirmacao":
        confirmacao?.toString() ?? "False", // Converte o bool para string
    "categoria":
        categoria ?? "Sem Categoria", // Valor padrão se categoria for nulo
  };
  return jsonData;
}

FFUploadedFile? listaArquivos(
  FFUploadedFile? inputImage1,
  FFUploadedFile? inputImage2,
  FFUploadedFile? inputImage3,
  FFUploadedFile? inputImage4,
  FFUploadedFile? inputImage5,
) {
  // Lista para consolidar os arquivos válidos
  List<FFUploadedFile> fileList = [];

  // Adiciona arquivos à lista caso não sejam nulos
  if (inputImage1 != null) {
    fileList.add(inputImage1);
  }
  if (inputImage2 != null) {
    fileList.add(inputImage2);
  }
  if (inputImage3 != null) {
    fileList.add(inputImage3);
  }
  if (inputImage4 != null) {
    fileList.add(inputImage4);
  }
  if (inputImage5 != null) {
    fileList.add(inputImage5);
  }

  // Retorna o primeiro arquivo da lista, ou null caso a lista esteja vazia
  return fileList.isNotEmpty ? fileList[0] : null;
}

String? formatDateTimeToDate(String? inputDateTime) {
  if (inputDateTime == null || inputDateTime.isEmpty) {
    return null; // Retorna nulo se a entrada for inválida
  }

  try {
    // Converte a string de entrada para um objeto DateTime
    DateTime parsedDateTime = DateTime.parse(inputDateTime);

    // Formata a data no formato dd/MM/yyyy
    final formattedDate = DateFormat('dd/MM/yyyy').format(parsedDateTime);

    return formattedDate;
  } catch (e) {
    // Retorna nulo em caso de erro de parsing
    return null;
  }
}

dynamic dependenteToJson(
  String? nomedependente,
  String? cpfdependente,
  String? parentescodependente,
  String? datanascimentodependente,
  String? id,
) {
  // Cria o mapa com os valores passados como argumentos
  final Map<String, dynamic> dependenteData = {
    "id": id,
    "nome": nomedependente,
    "cpf": cpfdependente,
    "parentesco": parentescodependente,
    "data_nascimento": datanascimentodependente,
  };

  // Retorna o mapa convertido para JSON
  return dependenteData;
}

int? convertStrToInt(String? str) {
  // Verifica se a String não é nula ou vazia
  if (str == null || str.isEmpty) {
    return null; // Retorna null se a entrada for inválida
  }

  try {
    // Tenta converter a String para um inteiro
    return int.parse(str);
  } catch (e) {
    // Retorna null em caso de erro (ex.: entrada inválida)
    return null;
  }
}

String? dateTimeToCurTime(String dateTimeString) {
  DateTime parsedDateTime = DateTime.parse(dateTimeString);
  DateTime now = DateTime.now();
  Duration difference = now.difference(parsedDateTime);

  if (difference.inMinutes < 60) {
    return '${difference.inMinutes} minutos atrás';
  } else if (difference.inHours < 24) {
    return '${difference.inHours} horas atrás';
  } else {
    return '${difference.inDays} dias atrás';
  }
}

List<int>? getNotificacaoIds(List<NotificacaoStruct> listaNotificacao) {
  return listaNotificacao.map((notificacao) => notificacao.id).toList();
}

String concatImagePaths(
  String imagePath1,
  String imagePath2,
) {
  if (imagePath1.isEmpty || imagePath2.isEmpty) {
    throw ArgumentError('Ambos os caminhos de imagem devem ser fornecidos.');
  }

  final concatenatedPath = "$imagePath1$imagePath2";

  return concatenatedPath;
}

String? versionBuild() {
  const String version = "1.0.0";
  const String buildNumber = "5";

  return 'Versão: $version, Build: $buildNumber';
}
